/*
 * FMDLog
 * Wideband O2 Display/Datalogging Application
 * For the TechEdge DIY-WB series of devices
 * (c) 2002-2003, Jonathan Burchmore
 *
 * Your use of this software is governed by the GNU General Public License, v2.
 * See the file COPYING.txt, or http://www.fsf.org/copyleft/gpl.txt
 *
 * This software is provided with ABSOLUTELY NO WARRANTY.  See the GNU General
 * public license for additional information.
 */

#include <PalmOS.h>
#include "resource.h"
#include "fmdlog.h"

/*
 * Alarm Functions (alarms.c)
 */

UInt32 alarm_min_RPM;
UInt8 alarm_afr, alarm_rpm, alarm_v1, alarm_v2;
double alarm_AFR_low, alarm_AFR_high;
UInt32 alarm_AFR_Hz,
	   alarm_RPM_low, alarm_RPM_high, alarm_RPM_Hz,
	   alarm_Vuser1_low, alarm_Vuser1_high, alarm_Vuser1_Hz,
	   alarm_Vuser2_low, alarm_Vuser2_high, alarm_Vuser2_Hz;

void Alarm_Play( int hz )
{
	SndCommandType cmd;

	cmd.cmd			= sndCmdFrqOn;
	cmd.reserved	= 0;
	cmd.param1		= hz;
	cmd.param2		= 1000;
	cmd.param3		= sndMaxAmp;

	SndDoCmd( NULL, &cmd, 0 );
}

Boolean AlarmForm_HandleEvent( EventPtr event )
{
	FormPtr frm;

	if ( event->eType == frmOpenEvent )
	{
		frm				= FrmGetActiveForm();

		Preferences_Set_Checkbox( frm,	checkboxAlarmAFR,	alarm_afr );
		Preferences_Set_Checkbox( frm,	checkboxAlarmRPM,	alarm_rpm );
		Preferences_Set_Checkbox( frm,	checkboxAlarmVu1,	alarm_v1 );
		Preferences_Set_Checkbox( frm,	checkboxAlarmVu2,	alarm_v2 );

		Preferences_Set_AFR( frm,		textAlarmAFRLow,	alarm_AFR_low );
		Preferences_Set_AFR( frm,		textAlarmAFRHigh,	alarm_AFR_high );
		Preferences_Set_Number( frm,	textAlarmAFRHz,		alarm_AFR_Hz );

		Preferences_Set_Number( frm,	textAlarmRPMLow,	alarm_RPM_low );
		Preferences_Set_Number( frm,	textAlarmRPMHigh,	alarm_RPM_high );
		Preferences_Set_Number( frm,	textAlarmRPMHz,		alarm_RPM_Hz );

		Preferences_Set_Number( frm,	textAlarmVu1Low,	alarm_Vuser1_low );
		Preferences_Set_Number( frm,	textAlarmVu1High,	alarm_Vuser1_high );
		Preferences_Set_Number( frm,	textAlarmVu1Hz,		alarm_Vuser1_Hz );

		Preferences_Set_Number( frm,	textAlarmVu2Low,	alarm_Vuser2_low );
		Preferences_Set_Number( frm,	textAlarmVu2High,	alarm_Vuser2_high );
		Preferences_Set_Number( frm,	textAlarmVu2Hz,		alarm_Vuser2_Hz );

		Preferences_Set_Number( frm,	textAlarmMinRPM,	alarm_min_RPM );

		FrmDrawForm( frm );
		return true;
	}
	else if ( event->eType == ctlSelectEvent )
	{
		if ( event->data.ctlSelect.controlID == buttonAlarmSave )
		{
			frm				= FrmGetActiveForm();

			Preferences_Get_Checkbox( frm,	checkboxAlarmAFR,	&alarm_afr );
			Preferences_Get_Checkbox( frm,	checkboxAlarmRPM,	&alarm_rpm );
			Preferences_Get_Checkbox( frm,	checkboxAlarmVu1,	&alarm_v1 );
			Preferences_Get_Checkbox( frm,	checkboxAlarmVu2,	&alarm_v2 );

			Preferences_Get_AFR( frm,		textAlarmAFRLow,	&alarm_AFR_low );
			Preferences_Get_AFR( frm,		textAlarmAFRHigh,	&alarm_AFR_high );
			Preferences_Get_NumberU32( frm,	textAlarmAFRHz,		&alarm_AFR_Hz );

			Preferences_Get_NumberU32( frm,	textAlarmRPMLow,	&alarm_RPM_low );
			Preferences_Get_NumberU32( frm,	textAlarmRPMHigh,	&alarm_RPM_high );
			Preferences_Get_NumberU32( frm,	textAlarmRPMHz,		&alarm_RPM_Hz );

			Preferences_Get_NumberU32( frm,	textAlarmVu1Low,	&alarm_Vuser1_low );
			Preferences_Get_NumberU32( frm,	textAlarmVu1High,	&alarm_Vuser1_high );
			Preferences_Get_NumberU32( frm,	textAlarmVu1Hz,		&alarm_Vuser1_Hz );

			Preferences_Get_NumberU32( frm,	textAlarmVu2Low,	&alarm_Vuser2_low );
			Preferences_Get_NumberU32( frm,	textAlarmVu2High,	&alarm_Vuser2_high );
			Preferences_Get_NumberU32( frm,	textAlarmVu2Hz,		&alarm_Vuser2_Hz );

			Preferences_Get_NumberU32( frm,	textAlarmMinRPM,	&alarm_min_RPM );

			Config_Save();
			FrmReturnToForm( 0 );
			return true;
		}
		else if ( event->data.ctlSelect.controlID == buttonAlarmCancel )
		{
			FrmReturnToForm( 0 );
			return true;
		}
	}

	return false;
}
