/*
 * FMDLog
 * Wideband O2 Display/Datalogging Application
 * For the TechEdge DIY-WB series of devices
 * (c) 2002-2003, Jonathan Burchmore
 *
 * Your use of this software is governed by the GNU General Public License, v2.
 * See the file COPYING.txt, or http://www.fsf.org/copyleft/gpl.txt
 *
 * This software is provided with ABSOLUTELY NO WARRANTY.  See the GNU General
 * public license for additional information.
 */

#include <PalmOS.h>
#include "resource.h"
#include "fmdlog.h"

/*
 * Edit Log Form Functions
 */

Boolean EditLogForm_HandleEvent( EventPtr event )
{
	FormPtr frm;
	FieldPtr field;
	UInt32 offset;
	Int16 selected;
	Char *recordp;
	MemHandle dbrecord, nameH, notesH;
	Char *name, *notes;
	ListPtr list;

	if ( event->eType == frmOpenEvent )
	{
		frm			= FrmGetActiveForm();

		if ( ( dbrecord = DmQueryRecord( dbP, dbrecord_index ) ) == NULL )
		{
			ErrFatalDisplay( "Unable to load logged data record" );
		}

		recordp		= MemHandleLock( dbrecord );
		offset		= Log_Read_Header( recordp, 0, &dbrecord_header );
		offset		= Log_Read_Desc( recordp, offset, &dbrecord_desc );

		FldSetMaxChars( FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, textEditLogName ) ),	sizeof( dbrecord_desc.name ) - 1 );
		FldSetMaxChars( FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, textEditLogNotes ) ),	sizeof( dbrecord_desc.notes ) - 1 );

		nameH		= MemHandleNew( sizeof( dbrecord_desc.name ) );
		notesH		= MemHandleNew( sizeof( dbrecord_desc.name ) );

		name		= MemHandleLock( nameH );
		notes		= MemHandleLock( notesH );

		StrNCopy( name,		dbrecord_desc.name,		sizeof( dbrecord_desc.name ) - 1 );
		StrNCopy( notes,	dbrecord_desc.notes,	sizeof( dbrecord_desc.notes ) - 1 );

		MemHandleUnlock( nameH );
		MemHandleUnlock( notesH );

		FldSetText( FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, textEditLogName ) ),	nameH,	0, sizeof( dbrecord_desc.name ) );
		FldSetText( FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, textEditLogNotes ) ),	notesH,	0, sizeof( dbrecord_desc.notes ) );

		FrmSetFocus( frm, FrmGetObjectIndex( frm, textEditLogNotes ) );
		FrmDrawForm( frm );

		DmReleaseRecord( dbP, dbrecord_index, 1 );
		return true;
	}
	else if ( event->eType == ctlSelectEvent )
	{
		if ( event->data.ctlSelect.controlID == buttonEditLogSave )
		{
			/*
			 * Update DB record with the notes from this form
			 */
			
			frm				= FrmGetActiveForm();
			field			= ( FieldPtr ) FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, textEditLogName ) );
			name			= FldGetTextPtr( field );
			field			= ( FieldPtr ) FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, textEditLogNotes ) );
			notes			= FldGetTextPtr( field );

			dbrecord		= DmGetRecord( dbP, dbrecord_index );
			recordp			= MemHandleLock( dbrecord );

			if ( name )		StrNCopy( dbrecord_desc.name,	name,	sizeof( dbrecord_desc.name ) - 1 );
			if ( notes )	StrNCopy( dbrecord_desc.notes,	notes,	sizeof( dbrecord_desc.notes ) - 1 );
			
			Log_Write_Desc( recordp, WBO2_RECORD_HEADER_SIZE, &dbrecord_desc );

			MemPtrUnlock( recordp );
			DmReleaseRecord( dbP, dbrecord_index, 1 );

			FrmReturnToForm( 0 );

			frm				= FrmGetActiveForm();
			list			= FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, listLoggedData ) );
			selected		= LstGetSelection( list );

			LoggedDataForm_LoadRecords( frm );
			LstSetSelection( list, selected );
			LstEraseList( list );
			LstDrawList( list );

			return true;
		}
		else if ( event->data.ctlSelect.controlID == buttonEditLogCancel )
		{
			FrmReturnToForm( 0 );
			return true;
		}
	}

	return false;
}

