/*
 * FMDLog
 * Wideband O2 Display/Datalogging Application
 * For the TechEdge DIY-WB series of devices
 * (c) 2002-2003, Jonathan Burchmore
 *
 * Your use of this software is governed by the GNU General Public License, v2.
 * See the file COPYING.txt, or http://www.fsf.org/copyleft/gpl.txt
 *
 * This software is provided with ABSOLUTELY NO WARRANTY.  See the GNU General
 * public license for additional information.
 */

#include <PalmOS.h>

#include "fmdlog.h"
#include "resource.h"
#include "resource-bitmaps.h"

#define AFR_DEFINE
#include "afr.h"
#include "afr2.h"

/*
 * Global Variables
 */

MemHandle			bitmap_handle_cache[ WBO2_BITMAP_MAX - WBO2_BITMAP_MIN + 1 ];
BitmapPtr			bitmap_ptr_cache[ WBO2_BITMAP_MAX - WBO2_BITMAP_MIN + 1 ];

UInt16				portID;
unsigned char		serial_buffer[ 8192 ];
unsigned char		packet_type = PF_TEv15;
int					synced = 0, packet_pos, packet_length = 12, packet_crc_pos = 11;
UInt32				Vwb, Vuser1, Vuser2, rpm, RPM;
UInt32				last_Vuser1, last_Vuser2, last_RPM;
int					lambda;
double				AFR, last_AFR;
unsigned char		packet[ 28 ];
unsigned int		sampleNo;
int					haveData, haveMarker, logActive;
int					drawDisplay;
Boolean				draw_paused = false;

int					graph_clear = 0;

graph_element		graph_afr, graph_rpm, graph_v1, graph_v2;
int					rpm_pulses;
UInt32				rpm_factor;

int					last_afr_y, last_rpm_y, last_v1_y, last_v2_y;
UInt32				ticks_lastdraw;
UInt16				ticks_second, ticks_slow;

#define AFR_SMOOTHING_N		7
double						afr_smoothing[ AFR_SMOOTHING_N ];
int							afr_smoothing_n, afr_smoothing_pos;

/*
 * Database Structures/Variables
 */

DmOpenRef				dbP;
WBO2Record_Header		dbrecord_header;
WBO2Record_Desc			dbrecord_desc;
unsigned char			dbrecord_queue[ 1024 ];
int						dbrecord_queue_pos, dbrecord_queue_samples;
UInt16					dbrecord_index, dbrecord_first_index;
UInt32					dbrecord_size;
UInt32					dbrecord_item_size;

/*
 * AFR Related Functions
 */

double AFR_Lookup( UInt32 v, AFRLookup *table, int n )
{
	int i;

	for ( i = n - 1; i >= 0; i-- )
	{
		if ( v >= table[ i ].Vout )	return table[ i ].AFR;
	}

	return 0.00;
}

/*
 * Main Form Functions
 */

Boolean MainForm_HandleEvent( EventPtr event )
{
	int i, alarming;
	double smoothed_afr;
	FormPtr frm;
	UInt32 ticks;
	RectangleType border;

	if ( event->eType == nilEvent && ( haveMarker || haveData ) )
	{
		if ( draw_paused )	return true;

		if ( haveMarker )
		{
			if ( logActive )
			{
				Log_Data( 1 );
			}

			/*
			 * Draw Display
			 */

			MainForm_DrawMarker();
		}

		if ( haveData )
		{
			switch ( packet_type )
			{
				case PF_TEv1 :
				{
					break;
				}
				case PF_TEv15 :
				{
					Vwb		= V_SCALE( ( ( packet[ 3 ] & 0x1F ) << 8 ) | packet[ 4 ] );
					Vuser1	= V_SCALE( ( ( packet[ 5 ] & 0x1F ) << 8 ) | packet[ 6 ] ); 
					Vuser2	= V_SCALE( ( ( packet[ 7 ] & 0x1F ) << 8 ) | packet[ 8 ] );
					rpm		= ( ( packet[ 9 ] & 0x1F ) << 8 ) | packet[ 10 ];

					if ( lambda )	AFR = AFR_Lookup( Vwb, lambda_lookup_table,	LAMBDA_TABLE_SIZE );
					else			AFR	= AFR_Lookup( Vwb, afr_lookup_table,	AFR_TABLE_SIZE );

					break;
				}
				case PF_TEv20 :
				{
					Vwb		= V2_VWB_SCALE( ( packet[ 5 ] & 0x1F ) << 8 ) | packet[ 6 ];
					Vuser1	= V_SCALE( ( ( packet[ 9 ] & 0x1F ) << 8 ) | packet[ 10 ] ); 
					Vuser2	= V_SCALE( ( ( packet[ 11 ] & 0x1F ) << 8 ) | packet[ 12 ] );
					rpm		= ( ( packet[ 23 ] & 0x1F ) << 8 ) | packet[ 24 ];

					if ( lambda )	AFR = AFR_Lookup( Vwb, lambda_lookup_table_v2,	LAMBDA_TABLE_V2_SIZE );
					else			AFR	= AFR_Lookup( Vwb, afr_lookup_table_v2,		AFR_TABLE_V2_SIZE );

					break;
				}
			}

			if ( scaling_v1_points )	Vuser1	= Sensor_Scale( Vuser1, scaling_v1_mv, scaling_v1_out, scaling_v1_points );
			if ( scaling_v2_points )	Vuser2	= Sensor_Scale( Vuser2, scaling_v2_mv, scaling_v2_out, scaling_v2_points );

			if ( rpm_pulses == 0 )
			{
				RPM	= rpm;
			}
			else
			{
				RPM	= rpm == 0 ? 0 : rpm_factor / rpm_pulses / rpm;
			}

			if ( RPM >= alarm_min_RPM )
			{
				alarming = 0;

				if ( !alarming && alarm_afr )
				{
					if ( AFR <= alarm_AFR_low || AFR >= alarm_AFR_high )
					{
						Alarm_Play( alarm_AFR_Hz );
						alarming	= 1;
					}
				}

				if ( !alarming && alarm_rpm )
				{
					if ( RPM <= alarm_RPM_low || RPM >= alarm_RPM_high )
					{
						Alarm_Play( alarm_RPM_Hz );
						alarming	= 1;
					}
				}

				if ( !alarming && alarm_v1 )
				{
					if ( Vuser1 <= alarm_Vuser1_low || Vuser1 >= alarm_Vuser1_high )
					{
						Alarm_Play( alarm_Vuser1_Hz );
						alarming	= 1;
					}
				}
				
				if ( !alarming && alarm_v2 )
				{
					if ( Vuser2 <= alarm_Vuser2_low || Vuser2 >= alarm_Vuser2_high )
					{
						Alarm_Play( alarm_Vuser2_Hz );
						alarming	= 1;
					}
				}
			}

			/*
			 * Log Data
			 */

			if ( logActive )
			{
				Log_Data( 0 );
			}

			/*
			 * Draw Display
			 */

			if ( graph_afr.display & ELEMENT_DISPLAY )
			{
				if ( graph_afr.display & ELEMENT_DISPLAY_SLOW )
				{
					ticks								= TimGetTicks();
					afr_smoothing[ afr_smoothing_pos ]	= AFR;

					if ( ++afr_smoothing_pos >= AFR_SMOOTHING_N )	afr_smoothing_pos = 0;
					if ( afr_smoothing_n < AFR_SMOOTHING_N )		afr_smoothing_n++;

					if ( ticks - ticks_lastdraw > ticks_slow )
					{
						ticks_lastdraw					= ticks;
						smoothed_afr					= 0.00;

						for ( i = 0; i < afr_smoothing_n; i++ )
						{
							smoothed_afr				+= afr_smoothing[ i ];
						}

						smoothed_afr					/= ( double ) i;
						smoothed_afr					+= 0.05;

						Draw_AFR( smoothed_afr, LCD_X, LCD_Y );
					}
				}
				else
				{
					Draw_AFR( AFR, LCD_X, LCD_Y );
				}
			}

			if ( graph_rpm.display )	Draw_RPM( RPM,	RPM_X,	RPM_Y );
			if ( graph_v1.display )		Draw_V( Vuser1,	U1_X,	U1_Y );
			if ( graph_v2.display )		Draw_V( Vuser2,	U2_X,	U2_Y );

			MainForm_DrawGraph();

			last_AFR	= AFR;
			last_Vuser1	= Vuser1;
			last_Vuser2	= Vuser2;
			last_RPM	= RPM;
		}

		haveData	= 0;
		haveMarker	= 0;

		return true;
	}
	else if ( event->eType == menuEvent )
	{
		switch ( event->data.menu.itemID )
		{
			case menuHardware :
			{
				FrmPopupForm( formHardware );
				break;
			}
			case menuPreferences :
			{
				FrmPopupForm( formPreferences );
				break;
			}
			case menuData :
			{
				if ( logActive )	FrmAlert( dialogEndLogError );
				else				FrmPopupForm( formLoggedData );

				break;
			}
			case menuLabels :
			{
				FrmPopupForm( formLabels );
				break;
			}
			case menuScalingVuser1 :
			{
				scaling_title	= "Sensor Scaling: Vuser1";
				scaling_mv		= scaling_v1_mv;
				scaling_out		= scaling_v1_out;
				scaling_points	= &scaling_v1_points;

				FrmPopupForm( formScaling );
				break;
			}
			case menuScalingVuser2 :
			{
				scaling_title	= "Sensor Scaling: Vuser2";
				scaling_mv		= scaling_v2_mv;
				scaling_out		= scaling_v2_out;
				scaling_points	= &scaling_v2_points;

				FrmPopupForm( formScaling );
				break;
			}
			case menuAlarms :
			{
				FrmPopupForm( formAlarms );
				break;
			}
			case menuColors :
			{
				FrmPopupForm( formColors );
				break;
			}
			case menuAbout :
			{
				FrmPopupForm( formAbout );
				break;
			}
		}

		return true;
	}
	else if ( event->eType == frmOpenEvent )
	{
		frm = FrmGetActiveForm();
		FrmDrawForm( frm );

		border.topLeft.x		= GRAPH_MIN_X;
		border.topLeft.y		= GRAPH_MIN_Y;
		border.extent.x			= GRAPH_WIDTH;
		border.extent.y			= GRAPH_HEIGHT;

		if ( graph_afr.display )	Draw_AFR( AFR, LCD_X, LCD_Y );
		if ( graph_rpm.display )	Draw_RPM( RPM,	RPM_X,	RPM_Y );

		if ( graph_v1.display )
		{
			Draw_V     ( Vuser1,	U1_X,	U1_Y );
			Draw_Vlabel( v1_label,	U1L_X,	U1L_Y );
		}

		if ( graph_v2.display )
		{
			Draw_V     ( Vuser2,	U2_X,	U2_Y );
			Draw_Vlabel( v2_label,	U2L_X,	U2L_Y );
		}

		WinDrawRectangleFrame( simpleFrame, &border );
		return true;
	}
	else if ( event->eType == winExitEvent )
	{
		if ( event->data.winExit.exitWindow == ( WinHandle ) FrmGetFormPtr( formMain ) )
		{
			draw_paused = true;
		}
	}
	else if ( event->eType == winEnterEvent )
	{
		if ( event->data.winEnter.enterWindow == ( WinHandle ) FrmGetFormPtr( formMain ) &&
			 event->data.winEnter.enterWindow == ( WinHandle ) FrmGetFirstForm() )
		{
			draw_paused = false;
		}
	}
	else if ( event->eType == keyDownEvent )
	{
		if ( event->data.keyDown.chr == vchrPageUp || event->data.keyDown.chr == vchrPageDown )
		{
			haveMarker	= 1;
			EvtWakeup();

			return true;
		}
		else if ( event->data.keyDown.chr == vchrHard1 ||
				  event->data.keyDown.chr == vchrHard2 ||
				  event->data.keyDown.chr == vchrHard3 ||
				  event->data.keyDown.chr == vchrHard4 )
		{
			if ( logActive )
			{
				Log_End();
			}
			else
			{
				Log_Start();
			}

			return true;
		}
	}

	return false;
}

void Draw_AFR( double afr, int x, int y )
{
	int digits[ 3 ];
	unsigned int data;
	static int bitmap_ids[] = { bitmapDigit0, bitmapDigit1, bitmapDigit2, bitmapDigit3, bitmapDigit4,
							    bitmapDigit5, bitmapDigit6, bitmapDigit7, bitmapDigit8, bitmapDigit9 };

	if ( lambda )
	{
		data					= afr * 100;
		digits[ 0 ]				= ( data % 1000 ) / 100;
		digits[ 1 ]				= ( data % 100 ) / 10;
		digits[ 2 ]				= data % 10;

		FMDLog_DrawBitmap( bitmap_ids[ digits[ 0 ] ],		x + 0,		y );
		FMDLog_DrawBitmap( bitmapDot,						x + 28,		y );
		FMDLog_DrawBitmap( bitmap_ids[ digits[ 1 ] ],		x + 38,		y );
		FMDLog_DrawBitmap( bitmap_ids[ digits[ 2 ] ],		x + 66,		y );
	}
	else
	{
		data					= afr * 10;
		digits[ 0 ]				= data / 100;
		digits[ 1 ]				= ( data % 100 ) / 10;
		digits[ 2 ]				= data % 10;

		FMDLog_DrawBitmap( bitmap_ids[ digits[ 0 ] ],		x + 0,		y );
		FMDLog_DrawBitmap( bitmap_ids[ digits[ 1 ] ],		x + 28,		y );
		FMDLog_DrawBitmap( bitmapDot,						x + 56,		y );
		FMDLog_DrawBitmap( bitmap_ids[ digits[ 2 ] ],		x + 66,		y );
	}
}

void Erase_AFR( int x, int y )
{
	RectangleType fillrect;

	fillrect.topLeft.x	= x;
	fillrect.topLeft.y	= y;
	fillrect.extent.x	= 94;
	fillrect.extent.y	= 48;

	WinEraseRectangle( &fillrect, 0 );
}

void Draw_RPM( unsigned int rpm, int x, int y )
{
	int digits[ 4 ];
	RectangleType fillrect;
	static int bitmap_ids[] = { bitmapDigit0t, bitmapDigit1t, bitmapDigit2t, bitmapDigit3t, bitmapDigit4t,
							    bitmapDigit5t, bitmapDigit6t, bitmapDigit7t, bitmapDigit8t, bitmapDigit9t };

	if ( rpm >= 10000 )
	{
		FMDLog_DrawBitmap( bitmapDigit10t,				x + 0,		y );
	}
	else
	{
		fillrect.topLeft.x	= x;
		fillrect.topLeft.y	= y;
		fillrect.extent.x	= 3;
		fillrect.extent.y	= 20;

		WinEraseRectangle( &fillrect, 0 );
	}

	digits[ 0 ]	= ( rpm % 10000 ) / 1000;
	digits[ 1 ]	= ( rpm % 1000 ) / 100;
	digits[ 2 ]	= ( rpm % 100 ) / 10;
	digits[ 3 ]	= rpm % 10;

	FMDLog_DrawBitmap( bitmap_ids[ digits[ 0 ] ],		x + 3,		y );
	FMDLog_DrawBitmap( bitmap_ids[ digits[ 1 ] ],		x + 18,		y );
	FMDLog_DrawBitmap( bitmap_ids[ digits[ 2 ] ],		x + 33,		y );
	FMDLog_DrawBitmap( bitmap_ids[ digits[ 3 ] ],		x + 48,		y );
}

void Erase_RPM( int x, int y )
{
	RectangleType fillrect;

	fillrect.topLeft.x	= x;
	fillrect.topLeft.y	= y;
	fillrect.extent.x	= 63;
	fillrect.extent.y	= 20;

	WinEraseRectangle( &fillrect, 0 );
}

void Draw_V( unsigned int V, int x, int y )
{
	int digits[ 4 ];
	static int bitmap_ids[] = { bitmapDigit0m, bitmapDigit1m, bitmapDigit2m, bitmapDigit3m, bitmapDigit4m,
							    bitmapDigit5m, bitmapDigit6m, bitmapDigit7m, bitmapDigit8m, bitmapDigit9m };

	if ( V >= 10000 )
	{
		digits[ 0 ]	= digits[ 1 ] = digits[ 2 ] = digits[ 3 ] = 9;
	}
	else
	{
		digits[ 0 ]	= ( V % 10000 ) / 1000;
		digits[ 1 ]	= ( V % 1000 ) / 100;
		digits[ 2 ]	= ( V % 100 ) / 10;
		digits[ 3 ]	= V % 10;
	}

	FMDLog_DrawBitmap( bitmap_ids[ digits[ 0 ] ],		x + 0,		y );
	FMDLog_DrawBitmap( bitmap_ids[ digits[ 1 ] ],		x + 8,		y );
	FMDLog_DrawBitmap( bitmap_ids[ digits[ 2 ] ],		x + 16,		y );
	FMDLog_DrawBitmap( bitmap_ids[ digits[ 3 ] ],		x + 24,		y );
}

void Erase_V( int x, int y )
{
	RectangleType fillrect;

	fillrect.topLeft.x	= x;
	fillrect.topLeft.y	= y;
	fillrect.extent.x	= 32;
	fillrect.extent.y	= 10;

	WinEraseRectangle( &fillrect, 0 );
}

void Draw_Vlabel( const char *label, int x, int y )
{
	int len;

	len	= StrLen( label );
	WinDrawChars( label, len, x - FntCharsWidth( label, len ), y );
}

void Erase_Vlabel( const char *label, int x, int y )
{
	int width;
	RectangleType fillrect;

	width				= FntCharsWidth( label, StrLen( label ) );
	fillrect.topLeft.x	= x - width;
	fillrect.topLeft.y	= y;
	fillrect.extent.x	= width;
	fillrect.extent.y	= FntCharHeight();

	WinEraseRectangle( &fillrect, 0 );
}

void FMDLog_DrawBitmap( int id, int x, int y )
{
	int index;

	index = id - WBO2_BITMAP_MIN;

	if ( bitmap_handle_cache[ index ] == 0 )	bitmap_handle_cache[ index ]	= DmGetResource( bitmapRsc, id );
	if ( bitmap_ptr_cache[ index ] == 0 )		bitmap_ptr_cache[ index ]		= MemHandleLock( bitmap_handle_cache[ index ] );

	WinDrawBitmap( bitmap_ptr_cache[ index ], x, y );
}

void Calculate_PPR( graph_element *graph, int height_px )
{
	graph->range		= graph->high - graph->low;
	graph->ppr			= ( ( double ) height_px ) / graph->range;
}

void Graph_Calculate_PPR()
{
	graph_clear		= 1;

	Calculate_PPR( &graph_afr, GRAPH_HEIGHT );
	Calculate_PPR( &graph_rpm, GRAPH_HEIGHT );
	Calculate_PPR( &graph_v1, GRAPH_HEIGHT );
	Calculate_PPR( &graph_v2, GRAPH_HEIGHT );
}

int MainForm_Graph_CalculateY( double data, graph_element *graph )
{
	if ( data <= graph->low )	return GRAPH_MAX_Y;
	if ( data >= graph->high )	return GRAPH_MIN_Y;

	return GRAPH_MIN_Y + ( GRAPH_HEIGHT - ( int ) ( graph->ppr * ( data - graph->low ) ) );
}

void MainForm_DrawGraph()
{
	int y;
	IndexedColorType last_color;

	RectangleType scroll, vacated;

	scroll.topLeft.x	= GRAPH_MIN_X;
	scroll.topLeft.y	= GRAPH_MIN_Y;
	scroll.extent.x		= GRAPH_WIDTH;
	scroll.extent.y		= GRAPH_HEIGHT;

	if ( graph_clear )
	{
		WinFillRectangle( &scroll, 0 );
		graph_clear = 0;
	}
	else
	{
		WinScrollRectangle( &scroll, winLeft, GRAPH_SAMPLE_WIDTH, &vacated );
		WinFillRectangle( &vacated, 0 );
	}

	/*
	 * Graph AFR
	 */

	if ( graph_afr.graph )
	{
		y			= MainForm_Graph_CalculateY( AFR, &graph_afr );
		last_color	= WinSetForeColor( color_afr );

		WinDrawLine( GRAPH_MIN_X + ( GRAPH_WIDTH - GRAPH_SAMPLE_WIDTH ),
					 last_afr_y,
					 GRAPH_MAX_X,
					 y );
		WinSetForeColor( last_color );

		last_afr_y	= y;
	}

	/*
	 * Graph Vuser1
	 */

	if ( graph_v1.graph )
	{
		y			= MainForm_Graph_CalculateY( Vuser1, &graph_v1 );
		last_color	= WinSetForeColor( color_v1 );

		WinDrawLine( GRAPH_MIN_X + ( GRAPH_WIDTH - GRAPH_SAMPLE_WIDTH ),
					 last_v1_y,
					 GRAPH_MAX_X,
					 y );
		WinSetForeColor( last_color );

		last_v1_y	= y;
	}

	/*
	 * Graph Vuser2
	 */

	if ( graph_v2.graph )
	{
		y			= MainForm_Graph_CalculateY( Vuser2, &graph_v2 );
		last_color	= WinSetForeColor( color_v2 );

		WinDrawLine( GRAPH_MIN_X + ( GRAPH_WIDTH - GRAPH_SAMPLE_WIDTH ),
					 last_v2_y,
					 GRAPH_MAX_X,
					 y );
		WinSetForeColor( last_color );

		last_v2_y	= y;
	}

	/*
	 * Graph RPM
	 */

	if ( graph_rpm.graph )
	{
		y			= MainForm_Graph_CalculateY( RPM, &graph_rpm );
		last_color	= WinSetForeColor( color_rpm );

		WinDrawLine( GRAPH_MIN_X + ( GRAPH_WIDTH - GRAPH_SAMPLE_WIDTH ),
					 last_rpm_y,
					 GRAPH_MAX_X,
					 y );
		WinSetForeColor( last_color );

		last_rpm_y	= y;
	}
}

void MainForm_DrawMarker()
{
	IndexedColorType color;
	RectangleType scroll, vacated;

	scroll.topLeft.x	= GRAPH_MIN_X;
	scroll.topLeft.y	= GRAPH_MIN_Y;
	scroll.extent.x		= GRAPH_WIDTH;
	scroll.extent.y		= GRAPH_HEIGHT;

	WinScrollRectangle( &scroll, winLeft, GRAPH_SAMPLE_WIDTH, &vacated );
	color = WinSetBackColor( UIColorGetTableEntryIndex( UIObjectForeground ) );
	WinFillRectangle( &vacated, 0 );
	WinSetBackColor( color );
}

/*
 * Logged Data Form Functions
 */

static Char		**loggeddata_items = 0;
static int		loggeddata_items_size = 0, loggeddata_items_pos = 0;

void LoggedData_Export()
{
	int i, data_pos, sample_pos, sample_no, samples_written;
	char sample_line[ 40 ], sample[ 5 ], temp[ 128 ]; 
	DmOpenRef memoDB;
	UInt16 memorecord_index;
	UInt32 offset, memorecord_size, memo_pos;
	UInt8 eAFR;
	UInt16 eVwb, eRPM, eVuser1, eVuser2;
	MemHandle recordH, memoH;
	Char *recordp, *memop;
	unsigned char *datap;

	if ( ( recordH = DmQueryRecord( dbP, dbrecord_index ) ) != NULL )
	{
		recordp				= MemHandleLock( recordH );
		offset				= Log_Read_Header( recordp, 0, &dbrecord_header );
		offset				= Log_Read_Desc( recordp, offset, &dbrecord_desc );
		datap				= ( unsigned char * ) recordp + offset;

		sample_no			= 1;
		samples_written		= 0;
		memorecord_index	= -1;
		memoDB				= DmOpenDatabaseByTypeCreator( 'DATA', 'memo', dmModeReadWrite );

		for ( ;; )
		{
			for ( i = 0, data_pos = 0; i < dbrecord_header.samples; i++ )
			{
				if ( samples_written == 0 )
				{
					if ( memorecord_index != -1 )
					{
						DmSet( memop, memo_pos, 1, 0 );
						MemHandleUnlock( memoH );
						DmReleaseRecord( memoDB, memorecord_index, 1 );
						DmResizeRecord( memoDB, memorecord_index, memo_pos +  1 );
					}

					memorecord_index	= dmMaxRecordIndex;
					memorecord_size		= StrLen( dbrecord_desc.name ) + 1 +
										  StrLen( dbrecord_desc.notes ) + 2 +
										  StrLen( "SAMPLE,Lambda,RPM,Vu1,Vu2\n" ) +
										  ( 150 * ( sizeof( sample_line ) ) ) + 1;
					memoH				= DmNewRecord( memoDB, &memorecord_index, memorecord_size );
					ErrFatalDisplayIf( memoH == NULL, "Unable to create memo record" );
					memop				= MemHandleLock( memoH );
					memo_pos			= 0;

					DmWrite( memop,	memo_pos, dbrecord_desc.name,		StrLen( dbrecord_desc.name ) );		memo_pos += StrLen( dbrecord_desc.name );
					DmWrite( memop, memo_pos, "\n",				1 );										memo_pos += 1;
					DmWrite( memop, memo_pos, dbrecord_desc.notes,		StrLen( dbrecord_desc.notes ) );	memo_pos += StrLen( dbrecord_desc.notes );
					DmWrite( memop, memo_pos, "\n\n",			2 );										memo_pos += 2;
					DmWrite( memop, memo_pos, "SAMPLE",			6 );										memo_pos += 6;

					if ( dbrecord_desc.fields & WBO2_FIELD_VWB )
					{
						if ( lambda )	{ DmWrite( memop, memo_pos, ",Lambda",	7 );		memo_pos += 7; }
						else			{ DmWrite( memop, memo_pos, ",AFR",		4 );		memo_pos += 4; }
					}

					if ( dbrecord_desc.fields & WBO2_FIELD_RPM )
					{
						DmWrite( memop, memo_pos, ",RPM",		4 );						memo_pos += 4;
					}

					if ( dbrecord_desc.fields & WBO2_FIELD_VU1 )
					{
						DmWrite( memop, memo_pos, ",Vu1",		4 );						memo_pos += 4;
					}

					if ( dbrecord_desc.fields & WBO2_FIELD_VU2 )
					{
						DmWrite( memop, memo_pos, ",Vu2",		4 );						memo_pos += 4;
					}

					DmWrite( memop, memo_pos, "\n",				1 );						memo_pos += 1;
				}

				eVwb	= 0;
				eRPM	= 0;
				eVuser1	= 0;
				eVuser2	= 0;
			
				if ( dbrecord_desc.fields & WBO2_FIELD_VWB ) { eVwb		= ( datap[ data_pos ] << 8 ) | datap[ data_pos + 1 ];	data_pos += 2; }
				if ( dbrecord_desc.fields & WBO2_FIELD_RPM ) { eRPM		= ( datap[ data_pos ] << 8 ) | datap[ data_pos + 1 ];	data_pos += 2; }
				if ( dbrecord_desc.fields & WBO2_FIELD_VU1 ) { eVuser1	= ( datap[ data_pos ] << 8 ) | datap[ data_pos + 1 ];	data_pos += 2; }
				if ( dbrecord_desc.fields & WBO2_FIELD_VU2 ) { eVuser2	 = ( datap[ data_pos ] << 8 ) | datap[ data_pos + 1 ];	data_pos += 2; }

				if ( eVwb == 0 && eRPM == 0 && eVuser1 == 0 && eVuser2 == 0 )
				{
					StrPrintF( sample_line, "%d,** Marker **\n", i + 1 );
				}
				else
				{
					switch ( packet_type )
					{
						case PF_TEv1 :
						{
							break;
						}
						case PF_TEv15 :
						{
							if ( lambda )	eAFR	= AFR_Lookup( eVwb, lambda_lookup_table,	LAMBDA_TABLE_SIZE );
							else			eAFR	= AFR_Lookup( eVwb, afr_lookup_table,		AFR_TABLE_SIZE );

							break;
						}
						case PF_TEv20 :
						{
							if ( lambda )	eAFR	= AFR_Lookup( eVwb, lambda_lookup_table_v2,		LAMBDA_TABLE_V2_SIZE );
							else			eAFR	= AFR_Lookup( eVwb, afr_lookup_table_v2,		AFR_TABLE_V2_SIZE );

							break;
						}
					}
					
					StrPrintF( sample_line, "%d", sample_no++ );

					if ( dbrecord_desc.fields & WBO2_FIELD_VWB )
					{
						if ( lambda )
						{
							sample_pos					= 0;

							sample[ sample_pos++ ]		= '0' + ( eAFR / 100 );
							sample[ sample_pos++ ]		= '.';
							sample[ sample_pos++ ]		= '0' + ( ( eAFR % 100 ) / 10 );
							sample[ sample_pos++ ]		= '0' + ( eAFR % 10 );
							sample[ sample_pos++ ]		= '\0';
						}
						else
						{
							sample_pos					= 0;

							if ( eAFR / 100 )
							{
								sample[ sample_pos++ ]	= '0' + ( eAFR / 100 );
							}

							sample[ sample_pos++ ]		= '0' + ( ( eAFR % 100 ) / 10 );
							sample[ sample_pos++ ]		= '.';
							sample[ sample_pos++ ]		= '0' + ( eAFR % 10 );
							sample[ sample_pos++ ]		= '\0';
						}

						StrCat( sample_line, "," );
						StrCat( sample_line, sample );
					}

					if ( dbrecord_desc.fields & WBO2_FIELD_RPM )
					{
						StrPrintF( temp, ",%u", eRPM );
						StrCat( sample_line, temp );
					}

					if ( dbrecord_desc.fields & WBO2_FIELD_VU1 )
					{
						StrPrintF( temp, ",%u", eVuser1 );
						StrCat( sample_line, temp );
					}

					if ( dbrecord_desc.fields & WBO2_FIELD_VU2 )
					{
						StrPrintF( temp, ",%u", eVuser2 );
						StrCat( sample_line, temp );
					}

					StrCat( sample_line, "\n" );
				}

				DmWrite( memop, memo_pos, sample_line,	StrLen( sample_line ) );	memo_pos += StrLen( sample_line );
				if ( ++samples_written >= 150 )	samples_written = 0;
			}

			MemHandleUnlock( recordH );
			DmReleaseRecord( dbP, dbrecord_index, 0 );

			if ( dbrecord_header.next == -1 )
			{
				if ( memorecord_index != -1 )
				{
					DmSet( memop, memo_pos, 1, 0 );
					MemHandleUnlock( memoH );
					DmReleaseRecord( memoDB, memorecord_index, 1 );
					DmResizeRecord( memoDB, memorecord_index, memo_pos +  1 );
				}

				break;
			}
			else
			{
				dbrecord_index	= dbrecord_header.next;
				if ( ( recordH = DmQueryRecord( dbP, dbrecord_index ) ) == NULL )	break;

				recordp			= MemHandleLock( recordH );
				offset			= Log_Read_Header( recordp, 0, &dbrecord_header );
				datap			= ( unsigned char * ) recordp + offset;
			}
		}

		DmCloseDatabase( memoDB );
	}
}

void LoggedDataForm_LoadRecords( FormPtr frm )
{
	ListPtr list;
	UInt32 offset;
	UInt16 dbrecord_index, i, numrecords, numitems;
	Char *recordp;
	WBO2Record_Header header;
	WBO2Record_Desc desc;
	MemHandle recordH;
	static Char *empty[] = { "(empty)" };

	list			= FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, listLoggedData ) );
	numrecords		= DmNumRecords( dbP );

	if ( numrecords == 0 )
	{
		LstSetListChoices( list, empty, 1 ); 
	}
	else
	{
		if ( loggeddata_items )
		{
			if ( loggeddata_items_size < numrecords )
			{
				for ( i = 0; i < loggeddata_items_pos; i++ )
				{
					MemPtrFree( loggeddata_items[ i ] );
				}

				MemPtrFree( loggeddata_items );
				loggeddata_items = NULL;
			}
		}

		if ( loggeddata_items == NULL )
		{
			loggeddata_items		= MemPtrNew( numrecords * sizeof( Char * ) );
			loggeddata_items_size	= numrecords;
			loggeddata_items_pos	= 0;
		}

		numitems		= 0;
		dbrecord_index	= 0;
		if ( !DmSeekRecordInCategory( dbP, &dbrecord_index, 0, dmSeekForward, 1 ) )
		{
			do
			{
				if ( ( recordH = DmQueryRecord( dbP, dbrecord_index ) ) != NULL )
				{
					recordp	= MemHandleLock( recordH );
					offset	= Log_Read_Header( recordp, 0, &header );
					offset	= Log_Read_Desc( recordp, offset, &desc );
					MemHandleUnlock( recordH );

					if ( numitems >= loggeddata_items_pos )
					{
						loggeddata_items[ numitems ]	= MemPtrNew( 30 + 1 );
						loggeddata_items_pos++;
					}

					StrNCopy( loggeddata_items[ numitems++ ], desc.name, 30 );
				}
			} while ( !DmSeekRecordInCategory( dbP, &dbrecord_index, 1, dmSeekForward, 1 ) );
		}

		LstSetListChoices( list, loggeddata_items, numitems ); 
	}
}

Boolean LoggedDataForm_HandleEvent( EventPtr event )
{
	int i;
	Int16 selected;
	FormPtr frm;
	ListPtr list;

	if ( event->eType == frmOpenEvent )
	{
		frm = FrmGetActiveForm();

		LoggedDataForm_LoadRecords( frm );
		FrmDrawForm( frm );

		return true;
	}
	else if ( event->eType == ctlSelectEvent )
	{
		if ( event->data.ctlSelect.controlID == buttonLoggedDataClose )
		{
			if ( loggeddata_items )
			{
				for ( i = 0; i < loggeddata_items_pos; i++ )
				{
					MemPtrFree( loggeddata_items[ i ] );
				}

				MemPtrFree( loggeddata_items );

				loggeddata_items		= 0;
				loggeddata_items_size	= 0;
				loggeddata_items_pos	= 0;
			}

			FrmReturnToForm( 0 );
			return true;
		}
		else if ( event->data.ctlSelect.controlID == buttonLoggedDataDelete ||
				  event->data.ctlSelect.controlID == buttonLoggedDataEdit ||
				  event->data.ctlSelect.controlID == buttonLoggedDataView ||
				  event->data.ctlSelect.controlID == buttonLoggedDataExport )
		{
			frm				= FrmGetActiveForm();
			list			= FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, listLoggedData ) );

			if ( ( selected = LstGetSelection( list ) ) == noListSelection )					goto error;

			dbrecord_index	= 0;
			if ( DmSeekRecordInCategory( dbP, &dbrecord_index, selected, dmSeekForward, 1 ) )	goto error;

			switch ( event->data.ctlSelect.controlID )
			{
				case buttonLoggedDataDelete :
				{
					Log_Delete_Chain( dbrecord_index );

					LoggedDataForm_LoadRecords( frm );
					LstEraseList( list );
					LstDrawList( list );

					break;
				}
				case buttonLoggedDataEdit :
				{
					FrmPopupForm( formEditLog );

					break;
				}
				case buttonLoggedDataExport :
				{
					LoggedData_Export();
					FrmAlert( dialogLoggedDataExported );

					break;
				}
				case buttonLoggedDataView :
				{
					FrmPopupForm( formLogReplay );
					break;
				}
			}

			return true;
error:
			FrmAlert( dialogLoggedDataError );
			return true;
		}
	}

	return false;
}

/*
 * Application Event Handler
 */

Boolean AppHandleEvent( EventPtr event )
{
	UInt16 formID;
	FormPtr frm;

	if ( event->eType == frmLoadEvent )
	{
		formID	= event->data.frmLoad.formID;
		frm		= FrmInitForm( formID );
		
		FrmSetActiveForm( frm );     

		/*
		 * Call the form initializer function, and set the event handler for the form.
		 */
		
		switch ( formID )
		{
			case formMain :
			{
				FrmSetEventHandler( frm, MainForm_HandleEvent );
				return true;
			}
			case formSaveLog :
			{
				FrmSetEventHandler( frm, SaveLogForm_HandleEvent );
				return true;
			}
			case formHardware :
			{
				FrmSetEventHandler( frm, HardwareForm_HandleEvent );
				return true;
			}
			case formPreferences :
			{
				FrmSetEventHandler( frm, PreferencesForm_HandleEvent );
				return true;
			}
			case formLoggedData :
			{
				FrmSetEventHandler( frm, LoggedDataForm_HandleEvent );
				return true;
			}
			case formEditLog :
			{
				FrmSetEventHandler( frm, EditLogForm_HandleEvent );
				return true;
			}
			case formLogReplay :
			{
				FrmSetEventHandler( frm, LogReplayForm_HandleEvent );
				return true;
			}
			case formScaling :
			{
				FrmSetEventHandler( frm, ScalingForm_HandleEvent );
				return true;
			}
			case formSmoothing :
			{
				FrmSetEventHandler( frm, SmoothingForm_HandleEvent );
				return true;
			}
			case formLabels :
			{
				FrmSetEventHandler( frm, LabelsForm_HandleEvent );
				return true;
			}
			case formAlarms :
			{
				FrmSetEventHandler( frm, AlarmForm_HandleEvent );
				return true;
			}
			case formAbout :
			{
				FrmSetEventHandler( frm, AboutForm_HandleEvent );
				return true;
			}
			case formColors :
			{
				FrmSetEventHandler( frm, ColorForm_HandleEvent );
				return true;
			}
		}
	}

	return false;
}

/*
 * Serial Port Wakeup Handler
 */

static void SerialRecvHandler( UInt32 refCon )
{
	int i;
	UInt32 crcraw, count, numread, numtoread, temp;
	UInt8 *bufP, crc;

	if ( SrmReceiveWindowOpen( portID, &bufP, &count ) != errNone )
	{
		SrmClearErr( portID );
	}
	else
	{
		numread	= 0;

		if ( haveData )
		{
			if ( count > 4096 )
			{
				numread	= count;
				synced	= 0;
			}

			EvtWakeup();
		}
		else if ( count )
		{
			if ( packet_type != PF_TEv1 && !synced )
			{
				if ( packet_pos == 0 )
				{
					while ( count )
					{
						count--;
						numread++;

						if ( *bufP++ == 0x5A )
						{
							packet[ packet_pos++ ]	= 0x5A;
							break;
						}
					}
				}

				if ( packet_pos == 1 && count )
				{
					if ( *bufP++ == 0xA5 )
					{
						count--;
						numread++;

						packet[ packet_pos++ ]	= 0xA5;
						synced					= 1;
					}
					else
					{
						packet_pos	= 0;
					}
				}
			}
			else
			{
				if ( count > packet_length )
				{
					/*
					 * Prevent overruns by dropping all except the last packet
					 */

					/*
					 * Drop any partial packet we may be holding
					 */

					if ( packet_pos )
					{
						temp		= packet_length - packet_pos;
						count		-= temp;
						numread		+= temp;
						bufP		+= temp;
					}

					temp	= ( ( count / packet_length ) - 1 ) * packet_length;
					bufP	+= temp;
					numread	+= temp;
					count	-= temp;
				}
			}

			numtoread	= packet_length - packet_pos > count ? count : packet_length - packet_pos;
			MemMove( &packet[ packet_pos ], bufP, numtoread );

			packet_pos	+= numtoread;
			numread		+= numtoread;

			if ( packet_pos == packet_length )
			{
				packet_pos	= 0;

				/*
				 * Verify CRC
				 */

				if ( packet_crc_pos == 0 )
				{
					haveData	= 1;
					EvtWakeup();
				}
				else
				{
if ( packet_length != 28 )	DbgMessage( "packet_length != 28" );
if ( packet_crc_pos != 27 )	DbgMessage( "packet_crc_pos != 27" );
					for ( crcraw = 0, i = 0; i < packet_length - 1; i++ )	crcraw += packet[ i ];
					crc			= ~crcraw;

					if ( crc != packet[ packet_crc_pos ] )
					{
						synced		= 0;
					}
					else
					{
						haveData	= 1;
						EvtWakeup();
					}
				}
			}
		}

		SrmReceiveWindowClose( portID, numread );
	}

	SrmSetWakeupHandler( portID, SerialRecvHandler, 0 );
	SrmPrimeWakeupHandler( portID, 1 );
}		

void DisableSerialHandler()
{
	SrmSetWakeupHandler( portID, NULL, 0 );
}

void EnableSerialHandler()
{
	SrmSetWakeupHandler( portID, SerialRecvHandler, 0 );
}

/*
 * CheckOSVersion
 */

int CheckOSVersion()
{
	UInt32 romversion;

	if ( FtrGet( sysFtrCreator, sysFtrNumROMVersion, &romversion ) )		return 0;
	if ( romversion < sysMakeROMVersion( 3, 3, 0, sysROMStageRelease, 0 ) )	return 0;

	return 1;
}

/*
 * PilotMain
 */

UInt32 PilotMain( UInt16 cmd, void *cmdPBP, UInt16 launchFlags )
{
	int i;
	Err error;
    EventType event;
	UInt32 flags;
	UInt16 flagsize;
	UInt16 autoofftime;
	UInt16 dbattr;
	LocalID dbid;

    if ( cmd == sysAppLaunchCmdNormalLaunch )
	{
		if ( !CheckOSVersion() )
		{
			FrmAlert( dialogOldOS );
		}
		else
		{
			last_AFR = AFR			= 0;
			last_Vuser1 = Vuser1	= 0;
			last_Vuser2 = Vuser2	= 0;
			last_RPM = RPM			= 0;

			last_afr_y				= GRAPH_MAX_Y;
			last_rpm_y				= GRAPH_MAX_Y;
			last_v1_y				= GRAPH_MAX_Y;
			last_v2_y				= GRAPH_MAX_Y;

			haveData				= 0;
			haveMarker				= 0;
			logActive				= 0;
			drawDisplay				= 1;
			autoofftime				= SysSetAutoOffTime( 0 );

			replayrecordH			= NULL;
			replayrecord_dbindex	= -1;

			ticks_lastdraw			= 0;
			ticks_second			= SysTicksPerSecond();
			ticks_slow				= ticks_second / 3;

			afr_smoothing_n			= 0;
			afr_smoothing_pos		= 0;

			MemSet( bitmap_handle_cache, 0, sizeof( MemHandle ) * ( WBO2_BITMAP_MAX - WBO2_BITMAP_MIN ) );
			MemSet( bitmap_handle_cache, 0, sizeof( BitmapPtr ) * ( WBO2_BITMAP_MAX - WBO2_BITMAP_MIN ) );

			Config_Load();
			Graph_Calculate_PPR();

			if ( ( dbP = DmOpenDatabaseByTypeCreator( WBO2_DBTYPE, WBO2_ID, dmModeReadWrite ) ) == 0 )
			{
				DmCreateDatabase( 0, WBO2_DBNAME, WBO2_ID, WBO2_DBTYPE, false );
				dbP = DmOpenDatabaseByTypeCreator( WBO2_DBTYPE, WBO2_ID, dmModeReadWrite );
			}

			if ( ( dbid = DmFindDatabase( 0, WBO2_DBNAME ) ) != 0 )
			{
				DmDatabaseInfo( 0, dbid, NULL, &dbattr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
				dbattr |= dmHdrAttrBackup;
				DmSetDatabaseInfo( 0, dbid, NULL, &dbattr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
			}

			if ( SrmOpen( 0x8000, 19200, &portID ) )
			{
				FrmAlert( dialogSerialSetupError );
			}
			else
			{
				flagsize = sizeof( flags );

				SrmSetReceiveBuffer( portID, serial_buffer, sizeof( serial_buffer ) );
				SrmControl( portID, srmCtlGetFlags, &flags, &flagsize );
				flags |= srmSettingsFlagFlowControl;
				flags |= srmSettingsFlagRTSAutoM;
				flags |= srmSettingsFlagCTSAutoM;
				SrmControl( portID, srmCtlSetFlags, &flags, &flagsize );
				
				SrmSetWakeupHandler( portID, SerialRecvHandler, 0 );
				SrmPrimeWakeupHandler( portID, 1 );
		
				FrmGotoForm( formMain );
				
				do
				{
					EvtGetEvent( &event, evtWaitForever );

					if ( event.eType == keyDownEvent &&
						 TxtCharIsHardKey( event.data.keyDown.modifiers, event.data.keyDown.chr ) )
					{
						if ( FrmDispatchEvent( &event ) )	continue;
					}
					
					if ( !SysHandleEvent( &event ) )
					{
						if ( !MenuHandleEvent( 0, &event, &error ) )
						{
							if ( !AppHandleEvent( &event ) )
							{
								FrmDispatchEvent( &event );
							}
						}
					}
				} while ( event.eType != appStopEvent );

				SrmSetReceiveBuffer( portID, NULL, 0 );
				SrmClose( portID );
				FrmCloseAllForms();
			}

			if ( loggeddata_items )
			{
				for ( i = 0; i < loggeddata_items_pos; i++ )
				{
					MemPtrFree( loggeddata_items[ i ] );
				}

				MemPtrFree( loggeddata_items );
			}

			for ( i = 0; i <= WBO2_BITMAP_MAX - WBO2_BITMAP_MIN; i++ )
			{
				if ( bitmap_ptr_cache[ i ] )	MemPtrUnlock( bitmap_ptr_cache[ i ] );
				if ( bitmap_handle_cache[ i ] )	DmReleaseResource( bitmap_handle_cache[ i ] );
			}

			if ( replayrecordH )				MemHandleUnlock( replayrecordH );
			if ( replayrecord_dbindex != -1 )	DmReleaseRecord( dbP, replayrecord_dbindex, 0 );
			if ( replaybitmap_window )			WinDeleteWindow( replaybitmap_window, false );
			if ( replay_title )					MemPtrFree( replay_title );

			DmCloseDatabase( dbP );
			SysSetAutoOffTime( autoofftime );
		}
    }

    return errNone;
}
