/*
 * FMDLog
 * Wideband O2 Display/Datalogging Application
 * For the TechEdge DIY-WB series of devices
 * (c) 2002-2003, Jonathan Burchmore
 *
 * Your use of this software is governed by the GNU General Public License, v2.
 * See the file COPYING.txt, or http://www.fsf.org/copyleft/gpl.txt
 *
 * This software is provided with ABSOLUTELY NO WARRANTY.  See the GNU General
 * public license for additional information.
 */

#include <PalmOS.h>
#include "resource.h"
#include "fmdlog.h"
#include "afr.h"
#include "afr2.h"

/*
 * Log Replay Form Functions
 */

BitmapPtr			replaybitmap;
WinHandle			replaybitmap_window;
MemHandle			replayrecordH;
UInt16				replayrecord_dbindex;
UInt32				replayrecord_samples, replayrecord_visible_samples;
WBO2Record_Header	*replayrecord_chain, *replayrecord_current;
WBO2Record_Desc		replayrecord_desc;
unsigned char		*replaydatap;
int					replay_item_size;
int					replaycurrent;
int					replayleftoffset;
int					replaypush;
int					replaypushamount;
int					replay_lastleftoffset;
int					replay_lastcurrent;
Char				*replay_title;

int					replay_afr_offset, replay_rpm_offset, replay_v1_offset, replay_v2_offset;

#define REPLAY_VWB( offset )	( ( replaydatap[ (offset) + replay_afr_offset ] << 8 ) | replaydatap[ (offset) + replay_afr_offset + 1 ] )
#define REPLAY_RPM( offset )	( ( replaydatap[ (offset) + replay_rpm_offset ] << 8 ) | replaydatap[ (offset) + replay_rpm_offset + 1 ] )
#define REPLAY_Vu1( offset )	( ( replaydatap[ (offset) + replay_v1_offset ] << 8 ) | replaydatap[ (offset) + replay_v1_offset + 1 ] )
#define REPLAY_Vu2( offset )	( ( replaydatap[ (offset) + replay_v2_offset ] << 8 ) | replaydatap[ (offset) + replay_v2_offset + 1 ] )

int LogReplayForm_Graph_CalculateY( double data, graph_element *graph )
{
	if ( data <= graph->low )	return GRAPH_MAX_Y;
	if ( data >= graph->high )	return GRAPH_MIN_Y;

	return ( GRAPH_HEIGHT - ( int ) ( graph->ppr * ( data - graph->low ) ) );
}

void LogReplayForm_Load( WBO2Record_Header *header )
{
	Char *recordp;

	if ( replayrecordH )				MemHandleUnlock( replayrecordH );
	if ( replayrecord_dbindex != -1 )	DmReleaseRecord( dbP, replayrecord_dbindex, 0 );

	replayrecord_current	= header;
	replayrecord_dbindex	= header->dbrecord_index;

	replayrecordH			= DmGetRecord( dbP, replayrecord_dbindex );
	if ( replayrecordH == NULL )	SysFatalAlert( "DmGetRecord in LogReplayForm_Load failed" );
	recordp					= MemHandleLock( replayrecordH );

	replaydatap				= ( unsigned char * ) recordp + header->data_offset;
}

void LogReplayForm_Data( int sample, double *rAFR, UInt16 *rRPM, UInt16 *rVuser1, UInt16 *rVuser2 )
{
	int data_pos;
	WBO2Record_Header *temp;

	if ( sample < replayrecord_current->first_sample )
	{
		temp		= replayrecord_current->prevp;
		while ( sample < temp->first_sample )
		{
			temp	= temp->prevp;
		}

		LogReplayForm_Load( temp );
	}
	else if ( sample > ( replayrecord_current->first_sample + replayrecord_current->samples - 1 ) )
	{
		temp		= replayrecord_current->nextp;
		while ( sample > ( temp->first_sample + temp->samples - 1 ) )
		{
			temp	= temp->nextp;
		}

		LogReplayForm_Load( temp );
	}

	data_pos	= ( sample - replayrecord_current->first_sample ) * replay_item_size;

	if ( replay_afr_offset != -1 )
	{
		switch ( packet_type )
		{
			case PF_TEv1 :
			{
				break;
			}
			case PF_TEv15 :
			{
				if ( lambda )	*rAFR	= AFR_Lookup( REPLAY_VWB( data_pos ), lambda_lookup_table,	LAMBDA_TABLE_SIZE );
				else			*rAFR	= AFR_Lookup( REPLAY_VWB( data_pos ), afr_lookup_table,		AFR_TABLE_SIZE );

				break;
			}
			case PF_TEv20 :
			{
				if ( lambda )	*rAFR	= AFR_Lookup( REPLAY_VWB( data_pos ), lambda_lookup_table_v2,	LAMBDA_TABLE_V2_SIZE );
				else			*rAFR	= AFR_Lookup( REPLAY_VWB( data_pos ), afr_lookup_table_v2,		AFR_TABLE_V2_SIZE );

				break;
			}
		}
	}

	if ( replay_rpm_offset != -1 )	*rRPM		= REPLAY_RPM( data_pos );
	if ( replay_v1_offset != -1 )	*rVuser1	= REPLAY_Vu1( data_pos );
	if ( replay_v2_offset != -1 )	*rVuser2	= REPLAY_Vu2( data_pos );
}

void LogReplayForm_Record( int sample, double *rAFR, UInt16 *rRPM, UInt16 *rVuser1, UInt16 *rVuser2 )
{
	double sAFR, tAFR;
	UInt32 sRPM, sVuser1, sVuser2;
	UInt16 tRPM, tVuser1, tVuser2;
	int i, n, afr_divisor, rpm_divisor, v1_divisor, v2_divisor;

	sample = sample + ( skip_factor * sample );
	if ( sample < 0 || sample >= replayrecord_samples )
	{
		*rAFR		= 0;
		*rRPM		= 0;
		*rVuser1	= 0;
		*rVuser2	= 0;
	}
	else
	{
		if ( !max_smoothing )
		{
			LogReplayForm_Data( sample, rAFR, rRPM, rVuser1, rVuser2 );
		}
		else
		{
			/*
			 * This code looks a little strange because I am attempting to
			 * read the data from lowest sample to highest sample to avoid
			 * swapping back and forth between database records when smoothing
			 * at a database record boundary
			 */

			sAFR		= 0;
			sRPM		= 0;
			sVuser1		= 0;
			sVuser2		= 0;

			afr_divisor	= 1;
			rpm_divisor	= 1;
			v1_divisor	= 1;
			v2_divisor	= 1;

			n = max_smoothing > sample ? sample : max_smoothing;
			for ( i = 1; i <= n; i++ )
			{
				LogReplayForm_Data( sample - i, &tAFR, &tRPM, &tVuser1, &tVuser2 );

				if ( smoothing_afr >= i )	{ sAFR		+= tAFR;	afr_divisor++; }
				if ( smoothing_rpm >= i )	{ sRPM		+= tRPM;	rpm_divisor++; }
				if ( smoothing_v1 >= i )	{ sVuser1	+= tVuser1;	v1_divisor++; }
				if ( smoothing_v2 >= i )	{ sVuser2	+= tVuser2;	v2_divisor++; }
			}

			LogReplayForm_Data( sample, &tAFR, &tRPM, &tVuser1, &tVuser2 );

			sAFR	+= tAFR;
			sRPM	+= tRPM;
			sVuser1	+= tVuser1;
			sVuser2	+= tVuser2;

			n = sample + max_smoothing > replayrecord_samples - 1 ? replayrecord_samples - sample - 1 : max_smoothing;
			for ( i = 1; i <= n; i++ )
			{
				LogReplayForm_Data( sample + i, &tAFR, &tRPM, &tVuser1, &tVuser2 );

				if ( smoothing_afr >= i )	{ sAFR		+= tAFR;	afr_divisor++; }
				if ( smoothing_rpm >= i )	{ sRPM		+= tRPM;	rpm_divisor++; }
				if ( smoothing_v1 >= i )	{ sVuser1	+= tVuser1;	v1_divisor++; }
				if ( smoothing_v2 >= i )	{ sVuser2	+= tVuser2;	v2_divisor++; }
			}

			*rAFR		= sAFR / afr_divisor;
			*rRPM		= sRPM / rpm_divisor;
			*rVuser1	= sVuser1 / v1_divisor;
			*rVuser2	= sVuser2 / v2_divisor;
		}
	}
}

void LogReplayForm_DrawGraphElement( int i, UInt16 xofs, LogReplayGraph_State *state )
{
	int j;
	RectangleType marker;
	IndexedColorType color;

	if ( state->from_AFR == 0 &&
		 state->from_RPM == 0 &&
		 state->from_Vuser1 == 0 &&
		 state->from_Vuser2 == 0 )
	{
		/*
		 * Draw Marker
		 */

		marker.topLeft.x	= xofs;
		marker.topLeft.y	= 0;
		marker.extent.x		= GRAPH_SAMPLE_WIDTH;
		marker.extent.y		= GRAPH_HEIGHT;

		color = WinSetBackColor( UIColorGetTableEntryIndex( UIObjectForeground ) );
		WinFillRectangle( &marker, 0 );
		WinSetBackColor( color );

		LogReplayForm_Record( i, &state->from_AFR, &state->from_RPM, &state->from_Vuser1, &state->from_Vuser2 );
	}
	else
	{
		LogReplayForm_Record( i, &state->to_AFR, &state->to_RPM, &state->to_Vuser1, &state->to_Vuser2 );

		if ( state->to_AFR == 0 &&
			 state->to_RPM == 0 &&
			 state->to_Vuser1 == 0 &&
			 state->to_Vuser2 == 0 )
		{
			for ( j = i; j < replayrecord_visible_samples; j++ )
			{
				LogReplayForm_Record( i, &state->to_AFR, &state->to_RPM, &state->to_Vuser1, &state->to_Vuser2 );

				if ( state->to_AFR || state->to_RPM || state->to_Vuser1 || state->to_Vuser2 )	break;
			}

			if ( state->to_AFR == 0 )		state->to_AFR		= state->from_AFR;
			if ( state->to_RPM == 0 )		state->to_RPM		= state->from_RPM;
			if ( state->to_Vuser1 == 0 )	state->to_Vuser1	= state->from_Vuser1;
			if ( state->to_Vuser2 == 0 )	state->to_Vuser2	= state->from_Vuser2;
		}

		/*
		 * Graph AFR
		 */

		if ( graph_afr.graph && ( replayrecord_desc.fields & WBO2_FIELD_VWB ) )
		{
			color = WinSetForeColor( color_afr );
			WinDrawLine( xofs,
						 LogReplayForm_Graph_CalculateY( state->from_AFR, &graph_afr ),
						 xofs + GRAPH_SAMPLE_WIDTH,
						 LogReplayForm_Graph_CalculateY( state->to_AFR, &graph_afr ) );
			WinSetForeColor( color );
		}

		/*
		 * Graph Vuser1
		 */

		if ( graph_v1.graph && ( replayrecord_desc.fields & WBO2_FIELD_VU1 ) )
		{
			color = WinSetForeColor( color_v1 );
			WinDrawLine( xofs,
						 LogReplayForm_Graph_CalculateY( state->from_Vuser1, &graph_v1 ),
						 xofs + GRAPH_SAMPLE_WIDTH,
						 LogReplayForm_Graph_CalculateY( state->to_Vuser1, &graph_v1 ) );
			WinSetForeColor( color );
		}

		/*
		 * Graph Vuser2
		 */

		if ( graph_v2.graph && ( replayrecord_desc.fields & WBO2_FIELD_VU2 ) )
		{
			color = WinSetForeColor( color_v2 );
			WinDrawLine( xofs,
						 LogReplayForm_Graph_CalculateY( state->from_Vuser2, &graph_v2 ),
						 xofs + GRAPH_SAMPLE_WIDTH,
						 LogReplayForm_Graph_CalculateY( state->to_Vuser2, &graph_v2 ) );
			WinSetForeColor( color );
		}

		/*
		 * Graph RPM
		 */

		if ( graph_rpm.graph && ( replayrecord_desc.fields & WBO2_FIELD_RPM ) )
		{
			color = WinSetForeColor( color_rpm );
			WinDrawLine( xofs,
						 LogReplayForm_Graph_CalculateY( state->from_RPM, &graph_rpm ),
						 xofs + GRAPH_SAMPLE_WIDTH,
						 LogReplayForm_Graph_CalculateY( state->to_RPM, &graph_rpm ) );
			WinSetForeColor( color );
		}
	
		state->from_AFR		= state->to_AFR;
		state->from_RPM		= state->to_RPM;
		state->from_Vuser1	= state->to_Vuser1;
		state->from_Vuser2	= state->to_Vuser2;
	}
}

void LogReplayForm_Draw( int full )
{
	int i;
	UInt16 xofs;
	RectangleType graphrect, scroll, vacated;
	WinHandle olddraw;
	LogReplayGraph_State state;

	LogReplayForm_Record( replaycurrent, &state.from_AFR, &state.from_RPM, &state.from_Vuser1, &state.from_Vuser2 );

	if ( graph_afr.display && ( replayrecord_desc.fields & WBO2_FIELD_VWB ) )
	{
		Draw_AFR( state.from_AFR, REPLAY_LCD_X, REPLAY_LCD_Y );
	}

	if ( graph_rpm.display && ( replayrecord_desc.fields & WBO2_FIELD_RPM ) )
	{
		Draw_RPM( state.from_RPM,	REPLAY_RPM_X,	REPLAY_RPM_Y );
	}

	if ( graph_v1.display && ( replayrecord_desc.fields & WBO2_FIELD_VU1 ) )
	{
					Draw_V     ( state.from_Vuser1,				REPLAY_U1_X,	REPLAY_U1_Y );
		if ( full )	Draw_Vlabel( replayrecord_desc.v1_label,	REPLAY_U1L_X,	REPLAY_U1L_Y );
	}

	if ( graph_v2.display && ( replayrecord_desc.fields & WBO2_FIELD_VU2 ) )
	{
					Draw_V     ( state.from_Vuser2,				REPLAY_U2_X,	REPLAY_U2_Y );
		if ( full )	Draw_Vlabel( replayrecord_desc.v2_label,	REPLAY_U2L_X,	REPLAY_U2L_Y );
	}

	olddraw	= WinSetDrawWindow( replaybitmap_window );
	if ( full )
	{
		/*
		 * Draw data into the replay bitmap
		 */

		graphrect.topLeft.x	= 0;
		graphrect.topLeft.y	= 0;
		graphrect.extent.x	= GRAPH_WIDTH;
		graphrect.extent.y	= GRAPH_HEIGHT;

		WinEraseRectangle( &graphrect, 0 );

		if ( replayleftoffset < replayrecord_visible_samples )
		{
			xofs		= 0;

			if ( replayleftoffset == 0 )	i = 0;
			else							i = replayleftoffset - 1;

			LogReplayForm_Record( i, &state.from_AFR, &state.from_RPM, &state.from_Vuser1, &state.from_Vuser2 );

			for ( i++;
				  i < replayrecord_visible_samples && i <= replayleftoffset + GRAPH_SAMPLES;
				  i++ )
			{
				LogReplayForm_DrawGraphElement( i, xofs, &state );

				xofs	+= GRAPH_SAMPLE_WIDTH;
			}
		}
	}
	else
	{
		/*
		 * Optimization: Scroll the bitmap just like the main form, instead of
		 *				 redrawing the entire thing
		 */

		/*
		 * Un-invert the last current
		 */

		graphrect.topLeft.x	= ( replay_lastcurrent - replay_lastleftoffset ) * GRAPH_SAMPLE_WIDTH;
		graphrect.topLeft.y	= 0;
		graphrect.extent.x	= GRAPH_SAMPLE_WIDTH;
		graphrect.extent.y	= GRAPH_HEIGHT;

		WinInvertRectangle( &graphrect, 0 );

		/*
		 * Scroll the bitmap
		 */

		scroll.topLeft.x	= 0;
		scroll.topLeft.y	= 0;
		scroll.extent.x		= GRAPH_WIDTH;
		scroll.extent.y		= GRAPH_HEIGHT;

		if ( replay_lastleftoffset < replayleftoffset )
		{
			WinScrollRectangle( &scroll, winLeft, GRAPH_SAMPLE_WIDTH * ( replayleftoffset - replay_lastleftoffset ), &vacated );
			WinFillRectangle( &vacated, 0 );

			/*
			 * Scrolling right, draw new data point(s)
			 */

			xofs		= GRAPH_WIDTH - ( GRAPH_SAMPLE_WIDTH * ( replayleftoffset - replay_lastleftoffset ) ) - 1;

			LogReplayForm_Record( replay_lastleftoffset + GRAPH_SAMPLES - 1,
								  &state.from_AFR, &state.from_RPM, &state.from_Vuser1, &state.from_Vuser2 );
			
			for ( i = replay_lastleftoffset + GRAPH_SAMPLES;
				  i <= replayleftoffset + GRAPH_SAMPLES;
				  i++ )
			{
				LogReplayForm_DrawGraphElement( i, xofs, &state );

				xofs	+= GRAPH_SAMPLE_WIDTH;
			}
		}
		else if ( replay_lastleftoffset > replayleftoffset )
		{
			WinScrollRectangle( &scroll, winRight, GRAPH_SAMPLE_WIDTH * ( replay_lastleftoffset - replayleftoffset ), &vacated );
			WinFillRectangle( &vacated, 0 );

			/*
			 * Scrolling left, draw new data point(s)
			 */

			LogReplayForm_Record( replayleftoffset, &state.from_AFR, &state.from_RPM, &state.from_Vuser1, &state.from_Vuser2 );

			xofs		= 0;

			for ( i = replayleftoffset + 1;
				  i <= replay_lastleftoffset; 
				  i++ )
			{
				LogReplayForm_DrawGraphElement( i, xofs, &state );

				xofs	+= GRAPH_SAMPLE_WIDTH;
			}
		}
	}

	/*
	 * Invert the area of the selected data point
	 */

	graphrect.topLeft.x	= ( replaycurrent - replayleftoffset ) * GRAPH_SAMPLE_WIDTH;
	graphrect.topLeft.y	= 0;
	graphrect.extent.x	= GRAPH_SAMPLE_WIDTH;
	graphrect.extent.y	= GRAPH_HEIGHT;

	WinInvertRectangle( &graphrect, 0 );

	WinSetDrawWindow( olddraw );

	/*
	 * Draw the bitmap to the window
	 */

	WinDrawBitmap( replaybitmap, GRAPH_MIN_X, GRAPH_MIN_Y );

	replay_lastleftoffset	= replayleftoffset;
	replay_lastcurrent		= replaycurrent;
}

Boolean LogReplayForm_HandleEvent( EventPtr event )
{
	Err error;
	FormPtr frm;
	int newcurrent;
	Int16 screenX, screenY;
	Boolean penDown;
	RectangleType border;

	if ( event->eType == frmOpenEvent )
	{
		frm								= FrmGetActiveForm();
		replayrecordH					= NULL;
		replayrecord_dbindex			= -1;

		Log_Read_Chain( dbrecord_index, &replayrecord_chain, &replayrecord_desc, &replayrecord_samples );
		LogReplayForm_Load( replayrecord_chain );

		replayrecord_visible_samples	= replayrecord_samples / ( skip_factor + 1 );

		replayleftoffset				= 0;
		replaycurrent					= 0;
		replaypush						= 0;

		replay_item_size				= 0;
		replay_afr_offset				= -1;
		replay_rpm_offset				= -1;
		replay_v1_offset				= -1;
		replay_v2_offset				= -1;

		if ( replayrecord_desc.fields & WBO2_FIELD_VWB ) { replay_afr_offset	= replay_item_size;	replay_item_size += 2; }
		if ( replayrecord_desc.fields & WBO2_FIELD_RPM ) { replay_rpm_offset	= replay_item_size;	replay_item_size += 2; }
		if ( replayrecord_desc.fields & WBO2_FIELD_VU1 ) { replay_v1_offset		= replay_item_size;	replay_item_size += 2; }
		if ( replayrecord_desc.fields & WBO2_FIELD_VU2 ) { replay_v2_offset		= replay_item_size;	replay_item_size += 2; }

		replaybitmap_window	= WinCreateOffscreenWindow( GRAPH_WIDTH, GRAPH_HEIGHT, screenFormat, &error );
		replaybitmap		= WinGetBitmap( replaybitmap_window );

		if ( replay_title == NULL ) replay_title = MemPtrNew( 20 + 1 );

		if ( StrLen( replayrecord_desc.name ) > 20 )
		{
			StrNCopy( replay_title, replayrecord_desc.name, 19 );

			replay_title[ 19 ]	= chrEllipsis;
			replay_title[ 20 ]	= '\0';
		}
		else
		{
			StrCopy( replay_title, replayrecord_desc.name );
		}

		FrmSetTitle( frm, replay_title );
		FrmDrawForm( frm );

		border.topLeft.x	= GRAPH_MIN_X;
		border.topLeft.y	= GRAPH_MIN_Y;
		border.extent.x		= GRAPH_WIDTH;
		border.extent.y		= GRAPH_HEIGHT;

		WinDrawRectangleFrame( simpleFrame, &border );
		LogReplayForm_Draw( 1 );

		return true;
	}
	else if ( event->eType == nilEvent )
	{
		if ( replaypush )
		{
			PenGetPoint( &screenX, &screenY, &penDown );
			if ( !penDown ||
			     ( replaypush < 0 && screenX > GRAPH_MIN_X + ( GRAPH_SAMPLE_WIDTH * 5 ) ) ||
				 ( replaypush > 1 && screenX < GRAPH_MIN_X + GRAPH_WIDTH - ( GRAPH_SAMPLE_WIDTH * 5 ) ) )
			{
				replaypush = 0;
			}
			else if ( replaypush < 0 )
			{
				/* Push Scroll Left */
				if ( replayleftoffset )
				{
					EvtSetNullEventTick( TimGetTicks() + ( SysTicksPerSecond() / 10 ) );

					replayleftoffset	-= replaypushamount;
					replaycurrent		= replayleftoffset + 5 - replaypushamount;

					if ( replayleftoffset < 0 )	replayleftoffset = 0;
					if ( replaycurrent < 0 )	replaycurrent = 0;

					LogReplayForm_Draw( 0 );
				}
			}
			else if ( replaypush > 0 )
			{
				/* Push Scroll Right */
				if ( replayleftoffset + GRAPH_SAMPLES < replayrecord_visible_samples )
				{
					/* Start push scroll right */
					EvtSetNullEventTick( TimGetTicks() + ( SysTicksPerSecond() / 10 ) );

					replayleftoffset	+= replaypushamount;
					replaycurrent		= replayleftoffset + GRAPH_SAMPLES - ( 5 - replaypushamount );

					if ( replayleftoffset > replayrecord_visible_samples - GRAPH_SAMPLES )	replayleftoffset = replayrecord_visible_samples - GRAPH_SAMPLES;
					if ( replaycurrent >= replayleftoffset + GRAPH_SAMPLES )				replaycurrent = replayleftoffset + GRAPH_SAMPLES - 1;

					LogReplayForm_Draw( 0 );
				}
			}
		}
	}
	else if ( event->eType == penDownEvent || event->eType == penMoveEvent )
	{
		if ( event->screenY >= GRAPH_MIN_Y && event->screenY < GRAPH_MIN_Y + GRAPH_HEIGHT )
		{
			if ( event->screenX <= GRAPH_MIN_X + ( GRAPH_SAMPLE_WIDTH * 5 ) && replayleftoffset )
			{
				/* Start push scroll left */
				replaypush			= -1;

				if ( event->screenX < GRAPH_MIN_X )
				{
					replaypushamount = 5;
				}
				else
				{
					replaypushamount	= 5 - ( ( event->screenX - GRAPH_MIN_X ) / GRAPH_SAMPLE_WIDTH );
					if ( replaypushamount <= 0 )	replaypushamount = 1;
				}

				EvtSetNullEventTick( TimGetTicks() + ( SysTicksPerSecond() / 10 ) );

				replayleftoffset	-= replaypushamount;
				replaycurrent		= replayleftoffset + 5 - replaypushamount;

				if ( replayleftoffset < 0 )	replayleftoffset = 0;
				if ( replaycurrent < 0 )	replaycurrent = 0;

				LogReplayForm_Draw( 0 );
			}
			else if ( event->screenX >= GRAPH_MIN_X + GRAPH_WIDTH - ( GRAPH_SAMPLE_WIDTH * 5 ) &&
					  replayleftoffset + GRAPH_SAMPLES < replayrecord_visible_samples )
			{
				/* Start push scroll right */
				replaypush			= 1;

				if ( event->screenX > GRAPH_MIN_X + GRAPH_WIDTH )
				{
					replaypushamount = 5;
				}
				else
				{
					replaypushamount	= 5 - ( ( ( GRAPH_MIN_X + GRAPH_WIDTH ) - event->screenX ) / GRAPH_SAMPLE_WIDTH );
					if ( replaypushamount <= 0 )	replaypushamount = 1;
				}

				EvtSetNullEventTick( TimGetTicks() + ( SysTicksPerSecond() / 10 ) );

				replayleftoffset	+= replaypushamount;
				replaycurrent		= replayleftoffset + GRAPH_SAMPLES - ( 5 - replaypushamount );

				if ( replayleftoffset > replayrecord_visible_samples - GRAPH_SAMPLES )	replayleftoffset = replayrecord_visible_samples - GRAPH_SAMPLES;
				if ( replaycurrent >= replayleftoffset + GRAPH_SAMPLES )				replaycurrent = replayleftoffset + GRAPH_SAMPLES - 1;

				LogReplayForm_Draw( 0 );
			}
			else
			{
				replaypush	= 0;
				newcurrent	= replayleftoffset + ( ( event->screenX - GRAPH_MIN_X ) / GRAPH_SAMPLE_WIDTH );

				if ( newcurrent < 0 )								newcurrent = 0;
				if ( newcurrent >= replayrecord_visible_samples )	newcurrent = replayrecord_visible_samples - 1;
				
				if ( newcurrent != replaycurrent )
				{
					replaycurrent = newcurrent;
					LogReplayForm_Draw( 0 );
				}
			}

			return true;
		}
	}
	else if ( event->eType == menuEvent )
	{
		switch ( event->data.menu.itemID )
		{
			case menuLogReplaySmoothing :
			{
				FrmPopupForm( formSmoothing );
				break;
			}
			case menuLogReplayExit :
			{
				WinDeleteWindow( replaybitmap_window, false );

				if ( replayrecordH )				MemHandleUnlock( replayrecordH );
				if ( replayrecord_dbindex != -1 )	DmReleaseRecord( dbP, replayrecord_dbindex, 0 );

				replayrecordH			= NULL;
				replayrecord_dbindex	= -1;
				replaybitmap			= NULL;
				replaybitmap_window		= NULL;

				Log_Free_Chain( &replayrecord_chain );
				FrmReturnToForm( 0 );

				break;
			}
		}

		return true;
	}

	return false;
}

