#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#define SCALE( x )			( ( x * 5000 ) / 8192 )
#define SCALE_WB( x )		( ( x * 5000 ) / 4096 )
#define RPM_PULSES_PER_REV	4

int Read( int fd, unsigned char *buffer, int size )
{
	int numread, totalread;

	for ( totalread = 0; totalread < size; totalread += numread )
	{
		if ( ( numread = read( fd, &buffer[ totalread ], size - totalread ) ) < 0 )	return -1;
	}

	return size;
}

typedef struct _AFRLookup
{
	int		Vout;
	double	AFR;
} AFRLookup;

#define AFR_DEFINE
#include "afr2.h"

double AFR_Lookup( unsigned int v, AFRLookup *table, int n )
{
	int i;

	for ( i = n - 1; i >= 0; i-- )
	{
		if ( v >= table[ i ].Vout )	return table[ i ].AFR;
	}

	return 0.00;
}

void main()
{
	int digits[ 3 ];
	unsigned int data;
	int i, synced, fd;
	unsigned int crcraw;
	unsigned int Tick, Vwb, Ipx, Vuser1, Vuser2, Vuser3, Thermo1, Thermo2, Thermo3, Therm, rpm, RPM, Status1, Status2;
	unsigned char packet[ 28 ], crc;

	synced	= 0;
	
	if ( ( fd = open( "/dev/ttyS0", O_RDONLY ) ) < 0 )	goto error;
	for ( ;; )
	{
		if ( synced )
		{
			if ( Read( fd, packet, 28 ) != 28 )			goto error;
		}
		else
		{
			printf( "." );	fflush( stdout );

			if ( read( fd, &packet[ 0 ], 1 ) != 1 )		goto error;
			if ( packet[ 0 ] != 0x5A )					continue;

			if ( read( fd, &packet[ 1 ], 1 ) != 1 )		goto error;
			if ( packet[ 1 ] != 0xA5 )					continue;

			synced = 1;
			printf( "Synchronized\n" );

			if ( Read( fd, &packet[ 2 ], 26 ) != 26 )	goto error;
		}

		printf( "crc = %u\n", ( unsigned int ) packet[ 27 ] );

		for ( crcraw = 0, i = 0; i < 27; i++ )	crcraw += packet[ i ];
		printf( "crcraw = %u, ~ = %u\n", ( unsigned int ) crcraw, ( unsigned int ) ( unsigned char ) ( ~crcraw ) );
		crc = ~crcraw;

		printf( "calculated crc = %u\n", ( unsigned int ) crc );

		if ( 0 && crc != packet[ 27 ] )
		{
			printf( "CRC Mismatch\n" );
			synced = 0;
		}
		else
		{
			Tick	= ( ( packet[ 3 ] & 0x1F ) << 8 ) | packet[ 4 ];
			Vwb		= ( ( packet[ 5 ] & 0x1F ) << 8 ) | packet[ 6 ];
			Ipx		= ( packet[ 7 ] << 8 ) | packet[ 8 ];
			Vuser1	= ( ( packet[ 9 ] & 0x1F ) << 8 ) | packet[ 10 ]; 
			Vuser2	= ( ( packet[ 11 ] & 0x1F ) << 8 ) | packet[ 12 ];
			Vuser3	= ( ( packet[ 13 ] & 0x1F ) << 8 ) | packet[ 14 ];
			Thermo1	= ( ( packet[ 15 ] & 0x1F ) << 8 ) | packet[ 16 ];
			Thermo2	= ( ( packet[ 17 ] & 0x1F ) << 8 ) | packet[ 18 ];
			Thermo3	= ( ( packet[ 19 ] & 0x1F ) << 8 ) | packet[ 20 ];
			Therm	= ( ( packet[ 21 ] & 0x1F ) << 8 ) | packet[ 22 ];
			rpm		= ( ( packet[ 23 ] & 0x1F ) << 8 ) | packet[ 24 ];
			Status1	= packet[ 25 ];
			Status2	= packet[ 26 ];

			if ( rpm == 0 )
			{
				RPM	= 0;
			}
			else
			{
				RPM	= ( ( 1000000.0 / 5.0 / rpm ) * 60 ) / RPM_PULSES_PER_REV;
				RPM	= 28000000 / RPM_PULSES_PER_REV / rpm;
			}

			data					= 14.7 * 10;
			digits[ 0 ]				= data / 100;
			digits[ 1 ]				= ( data % 100 ) / 10;
			digits[ 2 ]				= data % 10;

			printf( "Packet\n" );
			printf( "\tCounter: %d\n", packet[ 2 ] );
			printf( "\tTick   : %d\n", Tick );
			printf( "\tVwb    : %.3f %d\n", SCALE_WB( Vwb ) / 1000.0, Vwb );
			printf( "\tAFR    : %.2f\n", ( ( SCALE_WB( Vwb ) / 1000.0 ) * 2 ) + 9 );
			printf( "\tLookup : %.1f %d%d.%d\n", AFR_Lookup( SCALE_WB( Vwb ), afr_lookup_table_v2, AFR_TABLE_V2_SIZE ), digits[ 0 ], digits[ 1 ], digits[ 2 ] );
			printf( "\tIpx    : %d\n", Ipx );
			printf( "\tVuser1 : %d\n", SCALE( Vuser1 ) );
			printf( "\tVuser2 : %d\n", SCALE( Vuser2 ) );
			printf( "\tVuser3 : %d\n", SCALE( Vuser3 ) );
			printf( "\tThermo1: %d\n", Thermo1 );
			printf( "\tThermo2: %d\n", Thermo2 );
			printf( "\tThermo3: %d\n", Thermo3 );
			printf( "\tTherm  : %d\n", Therm );
			printf( "\trpm    : %d\n", rpm );
			printf( "\tRPM    : %d\n", RPM );
			printf( "\tStatus1: 0x%02x\n", Status1 );
			printf( "\tStatus2: 0x%02x\n", Status2 );
		}
	}

	return;

error:
	printf( "Error\n" );
}
